/*
 * GTK cryptmount
 * Copyright (c) 2013, Phillip Berndt
 *
 * Asks for a passphrase and mounts given LUKS partitions
 * Useful for singe-user systems upon GDM/KDM start.
 *
 * Licensed as GNU GPL v3.
 * See http://www.gnu.org/copyleft/gpl.html for license text.
 * Or ask me.
 */
#include <unistd.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <stdio.h>
#include <sys/stat.h>
#define _GNU_SOURCE
#include <string.h>
#include <libgen.h>
#include <dirent.h>
#include <libcryptsetup.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "gtksecentry.h"
#include "secmem/memory.h"

void checkParameters(int argc, char *argv[]) { /* {{{ */
	struct stat tester;
	DIR *testdir;

	/* Parameter count matches */
	if(argc != 3) {
		printf("gcryptmount - Mount LUKS partitions with graphical passphrase dialog\n"
			"\nUsage:\n gcryptmount <device> <mountpoint>\n"
			"\nWhere <device> is a block device and <mountpoint> an empty directory.\n\n"
			"If <mountpoint> is not empty this program will quit without any notice!\n");
		exit(1);
	}

	/* Must be run as root */
	if(getuid() != 0) {
		fprintf(stderr, "Error: This program must be run as root\n");
		exit(1);
	}

	/* 1st parameter is block device */
	if(stat(argv[1], &tester) != 0 || !S_ISBLK(tester.st_mode)) {
		fprintf(stderr, "Error: %s is not a valid block device\n", argv[1]);
		exit(1);
	}

	/* 2nd parameter is directory */
	if(stat(argv[2], &tester) != 0 || !S_ISDIR(tester.st_mode))
	{
		fprintf(stderr, "Error: %s is not a direcoty\n", argv[2]);
		exit(1);
	}

	/* Test if argv[2] is empty */
	if((testdir = opendir(argv[2])) != NULL) {
		readdir(testdir); /* Skip . and .. */
		readdir(testdir);
		if(readdir(testdir) != NULL) {
			closedir(testdir);

			/* Guess we already mounted this */
			exit(2);

		}
		closedir(testdir);
	}
	else {
		fprintf(stderr, "Error: Failed to open directory %s\n", argv[2]);
	}

} /* }}} */

gboolean pwEntryKeyEvent(GtkWidget *widget, GdkEventKey *event, gpointer dialog) {
	if(event->keyval == GDK_Return) {
		gtk_dialog_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
		return TRUE;
	}
	return FALSE;
}

void logfn(int level, const char *msg, void *usrptr) {
	printf("gcryptmount[%d]: %s\n", level, msg);
}

int main(int argc, char *argv[]) {
	GtkWidget *dialog;
	GtkWidget *pwEntry;
	GtkWidget *dialogIcon;
	GtkWidget *errLabel;
	char *luksDevName;
	char *errMessageFormatter;
	char errMessage[255];
	char *deviceFile;
	int dialogResponse;

	/* Check parameters */
	checkParameters(argc, argv);

	/* Initialize GTK */
	if(gtk_init_check(&argc, &argv) == FALSE) {
		fprintf(stderr, "Error: Failed to open X11 Display.\n");
		exit(1);
	}

	/* Compile a short version of <device> for LUKS naming */
	if((luksDevName = calloc(1, strlen(argv[1]) + 4)) == NULL) {
		fprintf(stderr, "Error: Failed to malloc space for device name\n");
		exit(1);
	}
	strcpy(&luksDevName[2], argv[1]);
	luksDevName = basename(&luksDevName[2]);
	luksDevName -= 2;
	luksDevName[0] = 'l';
	luksDevName[1] = '_';

	/* Initialize secure memory */
	secmem_init(1024);

	/* Create passphrase window */
	memset(errMessage, 0, 255);
	dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_QUESTION, GTK_BUTTONS_OK_CANCEL, "Please enter LUKS passphrase for %s", argv[2]);
	gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
	gtk_window_set_title(GTK_WINDOW(dialog), "gCryptMount");
	dialogIcon = gtk_image_new_from_stock(GTK_STOCK_DIALOG_AUTHENTICATION, GTK_ICON_SIZE_MENU);
	gtk_widget_show(dialogIcon);
	gtk_message_dialog_set_image(GTK_MESSAGE_DIALOG(dialog), dialogIcon);
	errLabel = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(gtk_dialog_get_content_area(GTK_DIALOG(dialog))), errLabel, FALSE, TRUE, 2);
	pwEntry = gtk_secure_entry_new();
	gtk_signal_connect(GTK_OBJECT(pwEntry), "key-press-event", G_CALLBACK(pwEntryKeyEvent), dialog);
	gtk_widget_show(pwEntry);
	gtk_box_pack_end(GTK_BOX(gtk_dialog_get_content_area(GTK_DIALOG(dialog))), pwEntry, FALSE, TRUE, 2);
	gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_CANCEL);

	struct crypt_device *luksDevice;
	if(crypt_init(&luksDevice, argv[1]) < 0) {
		printf("Failed to initialize crypt_device\n");
		exit(1);
	}
	crypt_memory_lock(luksDevice, 1);
	crypt_set_log_callback(luksDevice, &logfn, NULL);
	if(crypt_load(luksDevice, CRYPT_LUKS1, NULL) < 0) {
		printf("Failed to load crypt header\n");
		exit(1);
	}

	/* Ask for passphrase until opening/resuming succeeds */
	while(TRUE) {
		/* Ask for passphrase */
		gtk_secure_entry_set_text(GTK_SECURE_ENTRY(pwEntry), "");
		dialogResponse = gtk_dialog_run(GTK_DIALOG(dialog));

		if(dialogResponse != GTK_RESPONSE_OK) {
			/* Quit if the user canceled */
			exit(1);
		}

		/* Unlock partition */
		const char *passPhrase = gtk_secure_entry_get_text(GTK_SECURE_ENTRY(pwEntry));
		if(crypt_activate_by_passphrase(luksDevice, luksDevName, CRYPT_ANY_SLOT, passPhrase, strlen(passPhrase), 0) >= 0) {
			gtk_secure_entry_set_text(GTK_SECURE_ENTRY(pwEntry), "");
			break;
		}
		else {
			gtk_secure_entry_set_text(GTK_SECURE_ENTRY(pwEntry), "");
			crypt_get_error(errMessage, sizeof(errMessage));
			gtk_widget_show(errLabel);
			errMessageFormatter = g_markup_printf_escaped("<span color=\"red\">%s</span>", errMessage);
			gtk_label_set_markup(GTK_LABEL(errLabel), errMessageFormatter);
			g_free(errMessageFormatter);
		}
	}

	/* Try to mount the partition
	 * Use /bin/mount as it has nice guessing features and handles mtab
	 */
	if(fork() == 0) {
		deviceFile = malloc(strlen("/dev/mapper/") + strlen(luksDevName) + 2);
		sprintf(deviceFile, "/dev/mapper/%s", luksDevName);
		execl("/bin/mount", "/bin/mount", deviceFile, argv[2], NULL);
		fprintf(stderr, "Error: Failed to run mount: %s\n", strerror(errno));
		free(deviceFile);
		exit(1);
	}
	siginfo_t mountInfo;
	waitid(P_ALL, 0, &mountInfo, WEXITED);
	if(mountInfo.si_status != 0 || mountInfo.si_code != CLD_EXITED) {
		/* Close partition again */
		crypt_deactivate(luksDevice, luksDevName);
		/* Show error message */
		dialog = gtk_message_dialog_new(NULL, GTK_DIALOG_MODAL, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK, "Failed to mount partiton");
		gtk_window_set_position(GTK_WINDOW(dialog), GTK_WIN_POS_MOUSE);
		gtk_dialog_run(GTK_DIALOG(dialog));
		/* Exit */
		exit(1);
	}

	/* We're done */
	exit(0);


}
